package com.rrr.iidtest;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import org.apache.commons.compress.compressors.bzip2.BZip2CompressorOutputStream;
import org.apache.commons.compress.compressors.gzip.GzipCompressorOutputStream;
import org.apache.commons.compress.compressors.lzma.LZMACompressorOutputStream;
import org.apache.commons.compress.compressors.xz.XZCompressorOutputStream;
import org.apache.commons.math3.random.MersenneTwister;
import org.apache.commons.math3.random.RandomGenerator;

public class CompressorThread implements Runnable {

    private Thread t;
    private final List<Integer> statistics;
    private volatile boolean runningStatus;
    private volatile boolean shouldIDie;
    private final byte[] originalSample;
    private final int noPermutations;
    private final long rgenSeed;
    private final int Ti;

    public CompressorThread(int noPermutations, byte[] originalSample, 
            int Ti, long rgenSeed) {
        this.shouldIDie = false;
        this.originalSample = originalSample;
        this.noPermutations = noPermutations;
        this.statistics = new ArrayList<>();
        this.runningStatus = true;
        this.rgenSeed = rgenSeed;
        this.Ti = Ti;

    }

    @Override
    public void run() {
        System.out.println("Compressor " + Thread.currentThread().getName()
                + " started.");
                int bufferSize = (int) (originalSample.length * 4.3);  // Allow 30% headroom.
        ByteArrayOutputStream bos = new ByteArrayOutputStream(bufferSize);
        RandomGenerator rgen = new MersenneTwister(rgenSeed);
        for (int i = 0; i < 100_000; i++) {  // Hail Mary rgen warm up procedure!
            rgen.nextInt();
        }

        /* Shuffle multiple copies of the sample. */
        for (int p = 0; p < noPermutations; p++) {
            if (shouldIDie) {  // Should this thread be stopped?
                break;
            }
            /* Create a new copy so that we're not re-shuffling shuffled data. */
            byte[] pSample = Arrays.copyOf(originalSample, originalSample.length);
            for (int i = pSample.length - 1; i > 0; i--) {
                int randomPosition = rgen.nextInt(i + 1);
                byte temp = pSample[i];
                pSample[i] = pSample[randomPosition];
                pSample[randomPosition] = temp;
            }

            /* Perform the compression, 4 independent times. */
            try (BZip2CompressorOutputStream bzOut = new BZip2CompressorOutputStream(bos);
                    GzipCompressorOutputStream gzOut = new GzipCompressorOutputStream(bos);
                    XZCompressorOutputStream xzOut = new XZCompressorOutputStream(bos);
                    LZMACompressorOutputStream lzOut = new LZMACompressorOutputStream(bos)) {

                /* Compress the sample with four different compressors,
                   and concatenate the results, having cleared out the buffer. */
                bos.reset();
                bzOut.write(pSample);
                gzOut.write(pSample);
                xzOut.write(pSample);
                lzOut.write(pSample);
            } catch (IOException ex) {
                System.out.println(ex);
            }
            int T = bos.toByteArray().length;

            /* Store results and update user. */
            statistics.add(T);
            System.out.print("Compressor "
                    + Thread.currentThread().getName()
                    + " Permutation "
                    + (p + 1));
            System.out.println(", Normalised statistic = " + (double)T / Ti);
        }
        runningStatus = false;
    }

    public void start() {
        System.out.println("Starting a compressor thread...");
        if (t == null) {
            t = new Thread(this);
            t.start();
        }
    }

    public synchronized List<Integer> getStatistics() {
        return statistics;
    }

    public boolean isRunning() {
        return runningStatus;
    }

    public void die() {
        shouldIDie = true;
    }
}
