package com.rrr.iidtest;

import com.google.gson.stream.JsonWriter;
import java.io.ByteArrayOutputStream;
import java.io.FileWriter;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.security.SecureRandom;
import java.time.LocalDateTime;
import java.time.temporal.ChronoUnit;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Random;
import java.util.Set;
import java.util.concurrent.TimeUnit;
import org.apache.commons.compress.compressors.bzip2.BZip2CompressorOutputStream;
import org.apache.commons.compress.compressors.gzip.GzipCompressorOutputStream;
import org.apache.commons.compress.compressors.lzma.LZMACompressorOutputStream;
import org.apache.commons.compress.compressors.xz.XZCompressorOutputStream;

/**
 * IID test based on the concept that general purpose compression algorithms
 * exploit correlation to create smaller archives.
 *
 * From https://tukaani.org/xz/. "The core of the XZ Utils compression code is
 * based on LZMA SDK, but it has been modified quite a lot to be suitable for XZ
 * Utils. The primary compression algorithm is currently LZMA2, which is used
 * inside the .xz container format. With typical files, XZ Utils create 30 %
 * smaller output than gzip and 15 % smaller output than bzip2."
 *
 * This version automatically detects the number of CPU cores.
 */
public class IIDTester {

    private final int ALPHA;
    private final int NO_THREADS;
    private final int NO_CYCLES;
    private final Set<CompressorThread> allCompressors;
    private final int cores;

    public static void main(String[] args) {
        try {
            String inFile = args[0];
            String outFile = args[1];
            System.out.println("Starting IID test...");
            IIDTester t = new IIDTester();
            t.test(inFile, outFile);
        } catch (Exception ex) {
            showUsage(ex);
        }
    }

    private static void showUsage(Exception ex) {
        System.out.println("We have a problem...");
        System.out.println(ex);
        System.out.println("\n\nUsage is: java -jar \"iid-tester.jar\"  "
                + " infile  outfile");
        System.out.println("    infile:    The file of data samples for IID testing.");
        System.out.println("    outfile:   The name of the file holding the "
                + "\n               test results for display.");
        System.out.println("For example:- ");
        System.out.println("java -jar \"iid-tester.jar\"  /tmp/test_sample   /tmp/results.json  ");
        System.out.println("Then display results from file 'results.json' with our "
                + "java-iid-display.py program.");
        System.out.println("\nEnsure that you read the README.TXT file adjacent to the "
                + "\nthe \"lib\" folder when installing this program. It "
                + "\ncontains instructions for dealing with all necessary "
                + "\nlibrary files.");
        System.out.println("");
        System.out.println("");
    }

    public IIDTester() {
        /* Employ (n - 1) threads on an n core CPU to maintain 
           some responsiveness. */
        cores = Runtime.getRuntime().availableProcessors();
        if (cores > 1) {
            NO_THREADS = cores - 1;
        } else {
            NO_THREADS = 1;
        }
        allCompressors = new HashSet<>();
        NO_CYCLES = 10_000 / NO_THREADS + NO_THREADS;
        ALPHA = 100;  // This increases the strength of the test, at the expense of re-runs.
    }

    private void test(String inFile, String outFile) throws Exception {
        /* Calculate original un-shuffled statistic, Ti, as the baseline. */
        byte[] sample = Files.readAllBytes(Paths.get(inFile));
        int bufferSize = (int) (sample.length * 4.3);  // Allow 30% headroom.
        ByteArrayOutputStream bos = new ByteArrayOutputStream(bufferSize);
        try (BZip2CompressorOutputStream bzOut = new BZip2CompressorOutputStream(bos);
                GzipCompressorOutputStream gzOut = new GzipCompressorOutputStream(bos);
                XZCompressorOutputStream xzOut = new XZCompressorOutputStream(bos);
                LZMACompressorOutputStream lzOut = new LZMACompressorOutputStream(bos)) {
            bzOut.write(sample);
            gzOut.write(sample);
            xzOut.write(sample);
            lzOut.write(sample);
        }
        int Ti = bos.toByteArray().length;
        System.out.println("Baseline: " + Ti);

        /* Make well distributed seeds for the workers' rgens as 
           they're all spawned at ~same time, and the MT is partly 
           seeded from the system time. This is for cryptographic
           purposes after all :-) */
        Random mtSeeder = new SecureRandom();

        /* Start off threaded compressors. */
        for (int i = 0; i < NO_THREADS; i++) {
            CompressorThread ct = new CompressorThread(NO_CYCLES,
                    sample, Ti, mtSeeder.nextLong());
            ct.start();
            allCompressors.add(ct);
        }

        /*  Monitor threads till they finish.
            This is a dirty but safe & efficient technique.
            Also calculate expected time to completion of
            10,000 permutations.*/
        List<Integer> statistics;
        int rank = 0;
        LocalDateTime then = LocalDateTime.now();
        do {
            TimeUnit.SECONDS.sleep(10);
            statistics = getStatistics();
            rank = calculateRank(statistics, Ti);
            System.out.println("CURRENT RANK = "
                    + rank);
            int completedPermutations = statistics.size();
            if (completedPermutations > 10) {    // To prevent stupid estimates.
                LocalDateTime expectedCompletionDate
                        = expectedCompletionDate(then, completedPermutations);
                long remainingRunningTime = remainingRunningTime(expectedCompletionDate);
                System.out.println("Remaining running time to full completion = "
                        + remainingRunningTime + "s");
                System.out.println("Estimated full completion date/time = "
                        + expectedCompletionDate);
            }
        } while ((stillRunning()
                & (rank < ALPHA)));

        /* Kill any remaining threads. */
        allCompressors.forEach((ct) -> {
            ct.die();
        });

        /* Print summary. */
        System.out.println("\n----------------------------------");
        if (rank == 0) {
            System.out.println("Ranked as < 1/10,000");
            System.out.println("Totally FAILED permutation test.");
        } else {
            System.out.println("Ranked as > " + rank + "/10,000");
            if (rank >= ALPHA) {
                System.out.println("*** PASSED permutation test. "
                        + "There is no evidence that the data is not IID ***");
                System.out.println("Based on " + statistics.size()
                        + " permutations.");
            } else {
                System.out.println("FAILED permutation test.");
            }
        }

        /* Dump sorted results to a file. Writes a JSON file as:-
                rank:number
                Ti value:number
                other T test statistics:array of numbers
         */
        Collections.sort(statistics);
        try (JsonWriter writer = new JsonWriter(new FileWriter(outFile))) {
            writer.beginObject();
            writer.name("rank").value(rank);
            writer.name("ti").value(Ti);
            writer.name("statistics");
            writer.beginArray();
            for (int T : statistics) {
                writer.value(T);
            }
            writer.endArray();
            writer.endObject();
        }

        System.out.println("Test results in file: " + outFile);
        System.out.println("\n----------------------------------");
        LineChart_AWT.plot(statistics, Ti);
    }

    private List<Integer> getStatistics() {
        /* Collect up results (so far).  */
        List<Integer> statistics = new ArrayList<>();
        allCompressors.forEach((ct) -> {
            statistics.addAll(ct.getStatistics());
        });
        return statistics;
    }

    private int calculateRank(List<Integer> statistics, int T) {
        int rank = 0;
        rank = statistics.stream().filter((Ti) -> (T >= Ti)).map((_item) -> 1)
                .reduce(rank, Integer::sum);
        return rank;
    }

    private boolean stillRunning() {
        Boolean runStatus = false;
        for (CompressorThread ct : allCompressors) {
            if (ct.isRunning()) {
                runStatus = true;
            }
        }
        return runStatus;
    }

    private float permutationRate(LocalDateTime then, int completedPermutations) {
        LocalDateTime now = LocalDateTime.now();
        long elapsedTime = ChronoUnit.SECONDS.between(then, now);
        float permutationRate = ((float) elapsedTime) / completedPermutations;      // time (sec) per permutation
        return permutationRate;
    }

    private long remainingRunningTime(LocalDateTime future) {
        LocalDateTime now = LocalDateTime.now();
        long runningTime = ChronoUnit.SECONDS.between(now, future);
        return runningTime;   // seconds
    }

    private LocalDateTime expectedCompletionDate(LocalDateTime then, int completedPermutations) {
        float permutationRate = permutationRate(then, completedPermutations);
        float expectedOverallDuration = NO_CYCLES * NO_THREADS * permutationRate;     // seconds
        LocalDateTime estimatedCompletion = then.plusSeconds((long) expectedOverallDuration);
        return estimatedCompletion;
    }
}
